import ij.*;
import ij.process.*;
import ij.gui.*;
import java.io.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import ij.plugin.*;
import ij.plugin.frame.*;

/** Apply a LUT to a Roi'ed subset of a forground image and overlay on top of a Roi'ed background image, and add admonishments.
 *
 * @author Fred Damen <fred@damen.org>
 *
 * Version History
 * 2020-04-01 Released
 */
public class Colorizer_Overlay implements PlugIn, DialogListener {

   ImagePlus newimp = new ImagePlus();
   int[] ids = WindowManager.getIDList();
   GenericDialog gd = null;
   MultiLineLabel sw = null;

   public void run(String arg) {
      if (ids == null) {
         IJ.showStatus("Colorizer_Overlay: no open images.");
         return;
         }

      newimp.show();

      gd = new NonBlockingGenericDialog("Colorizer Overlay"); 

      Font font = gd.getFont();
      font = font.deriveFont(Font.BOLD, font.getSize2D()*1.5F);
      gd.addMessage("\nFirst...",font);
      Vector<String> itv = new Vector<String>();
      itv.add("<choose>");
      for(int id : ids)
         itv.add(WindowManager.getImage(id).getTitle());
      gd.addChoice("Background:", itv.toArray(new String[0]), "");
      gd.addChoice("Foreground:", itv.toArray(new String[0]), ""); 
      gd.addStringField("Range:", ""); gd.addToSameRow(); gd.addCheckbox("Query",false);
      statusMsg((Component)gd.getCheckboxes().get(0),"Determine range either when blank, or, always.");

      File lutsdir = new File(IJ.getDirectory("luts"));
      String[] luts = lutsdir.list(new FilenameFilter() {
	                       public boolean accept(File dir, String name) { return name.endsWith(".lut"); }
	                        });
      gd.addChoice("LUT:", concat("<none>",luts),"");

      gd.addMessage("\nthen...",font);

      gd.addStringField("Scale:","100%x100%");    gd.addToSameRow(); gd.addMessage("WxH;w,h");
      statusMsg(gd.getMessage(), "W width scaling factor\nH height scalling factor\nw added Left/Right pixels\nh added Top/Bottom pixels");
      gd.addChoice("Interpolation:",new String[]{"None","Bilinear","Bicubic"},"Bilinear");
      gd.addChoice("Boarder:",new String[]{"Top-Left", "Top-Center", "Top-Right", 
                                           "Center-Left", "Center", "Center-Right",
                                           "Bottom-Left", "Bottom-Center", "Bottom-Right"}, "Center");
                                                  gd.addToSameRow(); gd.addMessage("image goes...");
      gd.addMessage("\nthen...",font);

      gd.addChoice("ColorBar:", new String[]{"Left","Right", "Top", "Bottom","Seperate","None"}, "None");
      gd.addStringField("Size:", ""); gd.addToSameRow(); gd.addMessage("WxH;in,out");
      statusMsg(gd.getMessage(), "W width (normal)\nH height (parallel)\nin add proximal pixels\nout add distal pixels");

      gd.addStringField("Label:","");       gd.addToSameRow(); gd.addMessage("a)");
      statusMsg(gd.getMessage(), "String in top left of image.");
      gd.addStringField("Multiplier:","1"); gd.addToSameRow(); gd.addMessage("prescale labels");
      statusMsg(gd.getMessage(), "Say the pixels are in Kg and you want mg units.");
      gd.addStringField("Format:","%.2f");  gd.addToSameRow(); gd.addMessage("for range");
      statusMsg(gd.getMessage(), "Think prinf...");

      gd.addChoice("Font:", getFonts(), "");
      gd.addChoice("Style:", new String[]{"Plain", "Bold", "Italic", "Bold+Italic"}, "Plain");
      //gd.addChoice("Justification:", new String[]{"Left", "Center", "Right"}, "Center");
      gd.addChoice("Color:", Colors.colors, "white");
      gd.addChoice("Background:", concat("None",Colors.colors), "None");
      gd.addSlider("Size:", 4, 200, 16);
      //gd.addSlider("Angle:", -90, 90, 0);
      //gd.addCheckbox("Antialiased text", false);

      for(TextField tf : (Vector<TextField>)gd.getStringFields())
         tf.setColumns(40);

      gd.addMessage("Useful, an unuseful, messages go here.\nand here.");
      sw = (MultiLineLabel)gd.getMessage();

      gd.addDialogListener(this);
      gd.getButtons()[0].setEnabled(false);
      gd.showDialog();
      if (gd.wasCanceled())
         return;
      }

   private void statusMsg(Component c, final String str) {
      c.addMouseMotionListener(new MouseMotionListener() {
         @Override public void mouseDragged(MouseEvent e) {};
         @Override public void mouseMoved(MouseEvent e) { /* if (e.getID() == MouseEvent.MOUSE_ENTERED) */ status(str); };
         });
      }
   private void status(String str)  { status(str,0); }
   private void warning(String str) { status(str,1); }
   private void error(String str)   { status(str,2); }
   private void status(String str, int severity) {
      sw.setText(str);
      gd.pack();
      }

   private static String[] concat(String a, String[] b) {
      String[] res = new String[1+b.length];
      res[0] = a;
      for(int i=0; i<b.length; i++)
         res[i+1] = b[i];
      return res;
      }

   /* Applu LUT to Roi'ed DisplayRange'd forground image, overlay on Roi'ed background image and crop to bounding box.
    */
   public static ImagePlus doOverlay(ImagePlus bgimp, Roi bgroi, ImagePlus fgimp, Roi fgroi, double wmin, double wmax, String lut) {

      if (fgimp == null) {
         if (bgimp != null) {
            ImagePlus newimp = bgimp.duplicate();
            newimp.setRoi(bgroi);
            if (bgroi != null)
               IJ.run(newimp,"Clear Outside","slice");
            newimp.setImage(newimp.crop());
            IJ.run(newimp,"RGB Color", "");
            newimp.changes = false;
            newimp.setTitle("Colorized "+bgimp.getTitle());
            return newimp;
            }
         else return null;
         }

      if (bgimp == null) {
         Roi tmproi = fgimp.getRoi();
         fgimp.deleteRoi();
         bgimp = fgimp.duplicate();
         fgimp.setRoi(tmproi);
         IJ.run(bgimp, "Select All", "");
         IJ.run(bgimp, "Clear", "slice");
         }
      else {
         bgimp = bgimp.duplicate();
         bgimp.setRoi(bgroi);
         if (bgroi != null)
            IJ.run(bgimp,"Clear Outside","slice");
         }

      Roi tmproi = bgimp.getRoi();  // Can not trust duplicate not to crop
      bgimp.deleteRoi();
      ImagePlus newimp = bgimp.duplicate();
      IJ.run(newimp,"RGB Color", "");
      if (tmproi != null) bgimp.setRoi(tmproi);

      tmproi = fgimp.getRoi();  // Can not trust duplicate not to crop
      fgimp.deleteRoi();
      ImagePlus tmpimp = fgimp.duplicate();
      if (tmproi != null) fgimp.setRoi(tmproi);

      if (Double.isNaN(wmin)) wmin = fgimp.getDisplayRangeMin();
      if (Double.isNaN(wmax)) wmax = fgimp.getDisplayRangeMax();
      tmpimp.setDisplayRange(wmin,wmax);
      if ((lut != null) && !lut.equals("<none>"))
         IJ.run(tmpimp, lut.replaceAll(".lut$",""), "");
      tmpimp.setRoi(fgroi);
      tmpimp.copy();
      newimp.setRoi(fgroi);
      newimp.paste();

      newimp.deleteRoi();
      newimp.setImage(newimp.crop());  // paste dose not happen until image is copied
      newimp.setRoi(bgroi);
      newimp.setImage(newimp.crop());

      newimp.changes = false;

      newimp.setTitle("Colorized "+bgimp.getTitle());
      return newimp;
      }

   /** add label to imp at x,y using font and colors.
    */
   public static ImagePlus addLabel(ImagePlus imp, double x, double y, String label, Font font, Color fgc, Color bgc) {
      ImagePlus newimp = imp.duplicate();
      if ((label == null) || label.equals(""))
         return newimp;

      TextRoi roi = new TextRoi(x, y, label, font);
      Rectangle lB = roi.getBounds();
      roi.setLocation(lB.x + 2, lB.y + 2);
      Color tmpfgc = Toolbar.getForegroundColor();
      Color tmpbgc = Toolbar.getBackgroundColor();
      Toolbar.setForegroundColor(fgc);
      Toolbar.setBackgroundColor(bgc);
      newimp.setRoi(roi);
      IJ.run(newimp, "Draw", "slice");
      Toolbar.setForegroundColor(tmpfgc);
      Toolbar.setBackgroundColor(tmpbgc);
      newimp.deleteRoi();

      return newimp;
      }

   private static int parseSizeStr(String str, String def, int length) {
      if (str.equals(""))     str = def;
      if (str.endsWith("%"))  return (int)(length*Float.parseFloat(str.substring(0,str.length()-1))/100);
                         else return (int)Float.parseFloat(str);
      }
   private static int[] parseSize(String cbsize, int w, int h, String wdef, String hdef) {
      String[] cbab = cbsize.split("\\s*;\\s*",-1);
      String cbSize = (cbab.length > 0) ? cbab[0] : "";

      String[] cbss = cbSize.split("\\s*x\\s*",-1);
      int cbw = parseSizeStr((cbss.length > 0) ? cbss[0] : "", wdef, w);
      int cbh = parseSizeStr((cbss.length > 1) ? cbss[1] : "", hdef, h);

      String cbBoarder = (cbab.length > 1) ? cbab[1] : "";
      String[] cbio = cbBoarder.split("\\s*,\\s*",-1);
      int cbi = parseSizeStr((cbio.length > 0) ? cbio[0] : "", "0", w); // length???
      int cbo = parseSizeStr((cbio.length > 1) ? cbio[1] : "", "0", w);

      return new int[]{cbw, cbh, cbi, cbo};
      }

   /** add a properly adorned colorbar of at edge of the image.
    */
   public static ImagePlus addColorBar(ImagePlus imp, String lut, String cbwhere, String cbsize, double fgll, double fgul, String format, Font font, Color fgc, Color bgc) {
      ImagePlus newimp = imp.duplicate();
      if ((cbwhere == null) || cbwhere.equals("None") || (font == null))
         return newimp;

try {
      int w = newimp.getWidth();
      int h = newimp.getHeight();

      double scale = 1;

      Color tmpbgc = Toolbar.getBackgroundColor();

      switch(cbwhere) {
         case "Left": {
              int[] c = parseSize(cbsize,w,h,"10%","60%");
              ImagePlus cbimp = IJ.createImage("ColorBar", "32-bit ramp", c[1], c[0], 1);
              if ((lut != null) && !lut.equals("<none>"))
                 IJ.run(cbimp, lut.replaceAll(".lut$",""), "");
              IJ.run(cbimp, "Rotate 90 Degrees Left", "");
              cbimp.copy();
              if (bgc != null) Toolbar.setBackgroundColor(bgc);
              if ((c[0]+c[2]+c[3]) > 0) 
                 IJ.run(newimp, "Canvas Size...", "width="+(w+c[0]+c[2]+c[3])+" height="+h+" position=Center-Right zero");
              if (bgc != null) Toolbar.setBackgroundColor(tmpbgc);
              if (lut == null) return newimp;
              newimp.setRoi(c[3],h/2-c[1]/2,c[0],c[1]);
              newimp.paste();
              Rectangle cbrec = newimp.getRoi().getBounds();
              newimp.deleteRoi();
              newimp.setImage(newimp.crop());  // paste dose not happen until image is copied
              
              String uls = Double.isNaN(fgul) ? "NaN" : String.format(format, fgul);
              String lls = Double.isNaN(fgll) ? "NaN" : String.format(format, fgll);
              //drawString(newimp, uls, cbrec, "Left", "Top",    font, fgc, bgc);
              //drawString(newimp, lls, cbrec, "Left", "Bottom", font, fgc, bgc);
              drawString(newimp, uls, cbrec, "LULU", font, fgc, bgc);
              drawString(newimp, lls, cbrec, "LLLL", font, fgc, bgc);
              return newimp;
              }
         case "Right": {
              int[] c = parseSize(cbsize,w,h,"10%","60%");
              ImagePlus cbimp = IJ.createImage("ColorBar", "32-bit ramp", c[1], c[0], 1);
              if ((lut != null) && !lut.equals("<none>"))
                 IJ.run(cbimp, lut.replaceAll(".lut$",""), "");
              IJ.run(cbimp, "Rotate 90 Degrees Left", "");
              cbimp.copy();
              if (bgc != null) Toolbar.setBackgroundColor(bgc);
              if ((c[0]+c[2]+c[3]) > 0) 
                 IJ.run(newimp, "Canvas Size...", "width="+(w+c[0]+c[2]+c[3])+" height="+h+" position=Center-Left");
              if (bgc != null) Toolbar.setBackgroundColor(tmpbgc);
              if (lut == null) return newimp;
              newimp.setRoi(w+c[2],h/2-c[1]/2,c[0],c[1]);
              newimp.paste();
              Rectangle cbrec = newimp.getRoi().getBounds();
              newimp.deleteRoi();
              newimp.setImage(newimp.crop());  // paste dose not happen until image is copied

              String uls = Double.isNaN(fgul) ? "NaN" : String.format(format, fgul);
              String lls = Double.isNaN(fgll) ? "NaN" : String.format(format, fgll);
              //drawString(newimp, uls, cbrec, "Right", "Top",    font, fgc, bgc);
              //drawString(newimp, lls, cbrec, "Right", "Bottom", font, fgc, bgc);
              drawString(newimp, uls, cbrec, "RURU", font, fgc, bgc);
              drawString(newimp, lls, cbrec, "RLRL", font, fgc, bgc);
              return newimp;
              }
         case "Top": {
              int[] c = parseSize(cbsize,h,w,"10%","60%");
              ImagePlus cbimp = IJ.createImage("ColorBar", "32-bit ramp", c[1], c[0], 1);
              if ((lut != null) && !lut.equals("<none>"))
                 IJ.run(cbimp, lut.replaceAll(".lut$",""), "");
              cbimp.copy();
              if (bgc != null) Toolbar.setBackgroundColor(bgc);
              if ((c[0]+c[2]+c[3]) > 0) 
                 IJ.run(newimp, "Canvas Size...", "width="+w+" height="+(h+c[0]+c[2]+c[3])+" position=Bottom-Center zero");
              if (bgc != null) Toolbar.setBackgroundColor(tmpbgc);
              if (lut == null) return newimp;
              newimp.setRoi(w/2-c[1]/2,c[3],c[1],c[0]);
              newimp.paste();
              Rectangle cbrec = newimp.getRoi().getBounds();
              newimp.deleteRoi();
              newimp.setImage(newimp.crop());  // paste dose not happen until image is copied

              String uls = Double.isNaN(fgul) ? "NaN" : String.format(format, fgul);
              String lls = Double.isNaN(fgll) ? "NaN" : String.format(format, fgll);
              //drawString(newimp, uls, cbrec, "Left", "Center", font, fgc, bgc);
              //drawString(newimp, lls, cbrec, "Right",  "Center", font, fgc, bgc);
              drawString(newimp, uls, cbrec, "LCRC", font, fgc, bgc);
              drawString(newimp, lls, cbrec, "RCLC", font, fgc, bgc);
              return newimp;
              }
         case "Bottom": {
              int[] c = parseSize(cbsize,h,w,"10%","60%");
              ImagePlus cbimp = IJ.createImage("ColorBar", "32-bit ramp", c[1], c[0], 1);
              if ((lut != null) && !lut.equals("<none>"))
                 IJ.run(cbimp, lut.replaceAll(".lut$",""), "");
              cbimp.copy();
              if (bgc != null) Toolbar.setBackgroundColor(bgc);
              if ((c[0]+c[2]+c[3]) > 0) 
                 IJ.run(newimp, "Canvas Size...", "width="+w+" height="+(h+c[0]+c[2]+c[3])+" position=Top-Center zero");
              if (bgc != null) Toolbar.setBackgroundColor(tmpbgc);
              if (lut == null) return newimp;
              newimp.setRoi(w/2-c[1]/2,h+c[2],c[1],c[0]);
              newimp.paste();
              Rectangle cbrec = newimp.getRoi().getBounds();
              newimp.deleteRoi();
              newimp.setImage(newimp.crop());  // paste dose not happen until image is copied

              String uls = Double.isNaN(fgul) ? "NaN" : String.format(format, fgul);
              String lls = Double.isNaN(fgll) ? "NaN" : String.format(format, fgll);
              //drawString(newimp, uls, cbrec, "Left", "Center", font, fgc, bgc);
              //drawString(newimp, lls, cbrec, "Right",  "Center", font, fgc, bgc);
              drawString(newimp, uls, cbrec, "LCRC", font, fgc, bgc);
              drawString(newimp, lls, cbrec, "RCLC", font, fgc, bgc);
              return newimp;
              }
         case "Seperate": {
              int[] c = parseSize(cbsize,w,h,"10%","60%");
              ImagePlus cbimp = IJ.createImage("ColorBar", "32-bit ramp", c[1], c[0], 1);
              if ((lut != null) && !lut.equals("<none>"))
                 IJ.run(cbimp, lut.replaceAll(".lut$",""), "");
              return cbimp;
              }
         case "none":
              return null;
         default:
              IJ.error("Should not have gottom here");
         }

      }
      catch(Exception e) {
         StringWriter sw = new StringWriter();
         e.printStackTrace(new PrintWriter(sw));
         IJ.error("An error occurred while trying to create a colorbar.\n\n"+sw.toString());
         }
      return null;
      }

   /** untility method to adorn the colorbar.
    */
   public static void drawString(ImagePlus imp, String str, Rectangle box, String where, Font font, Color fgc, Color bgc) {
      TextRoi roi = new TextRoi(box.x+box.width/2, box.y+box.height/2, str, font);
      Rectangle tb = roi.getBounds();

      switch(where.concat("    ").substring(0,0+1)) {
      case "R": tb.x += box.width/2; break;
      case "C":                      break;
      case "L": tb.x -= box.width/2; break;
      }
      switch(where.concat("    ").substring(1,1+1)) {
      case "L": tb.y += box.height/2; break;
      case "C":                       break;
      case "U": tb.y -= box.height/2; break;
      }
      switch(where.concat("    ").substring(2,2+1)) {
      case "R": tb.x -= tb.width/2;
      case "C": tb.x -= tb.width/2;
      case "L": 
      }
      tb.y -= tb.height/2;
      switch(where.concat("    ").substring(3,3+1)) {
      case "L": tb.y += tb.height/2; break;
      case "C":                      break;
      case "U": tb.y -= tb.height/2; break;
      }

      roi.setLocation(tb.x, tb.y);
      Color tmpfgc = Toolbar.getForegroundColor();
      Color tmpbgc = Toolbar.getBackgroundColor();
      Toolbar.setForegroundColor(fgc);
      Toolbar.setBackgroundColor(bgc);
      imp.setRoi(roi);
      IJ.run(imp, "Draw", "slice");
      Toolbar.setForegroundColor(tmpfgc);
      Toolbar.setBackgroundColor(tmpbgc);
      imp.deleteRoi();
      }

   private static void drawString(ImagePlus imp, String str, Rectangle box, String hj, String vj, Font font, Color fgc, Color bgc) {
      TextRoi roi = new TextRoi(box.x, box.y, str, font);
      Rectangle tb = roi.getBounds();
      switch(vj) {
         case "Top":    tb.y = box.y+box.height;               break;
         case "Center": tb.y = box.y+(box.height-tb.height)/2; break;
         case "Bottom": tb.y = box.y-tb.height;                break;
         default:  IJ.error("Invalid vertical justification");
         }
      switch(hj) {
         case "Left":   tb.x = box.x;                        break;
         case "Center": tb.x = box.x+(box.width-tb.width)/2; break;
         case "Right":  tb.x = box.x+box.width-tb.width;     break;
         default:  IJ.error("Invalid horizontal justification");
         }
      roi.setLocation(tb.x, tb.y);
      Color tmpfgc = Toolbar.getForegroundColor();
      Color tmpbgc = Toolbar.getBackgroundColor();
      Toolbar.setForegroundColor(fgc);
      Toolbar.setBackgroundColor(bgc);
      imp.setRoi(roi);
      IJ.run(imp, "Draw", "slice");
      Toolbar.setForegroundColor(tmpfgc);
      Toolbar.setBackgroundColor(tmpbgc);
      imp.deleteRoi();
      }

   public boolean dialogItemChanged(GenericDialog gd, AWTEvent ev) {
      if (ev==null) return false;
      if (ev instanceof java.awt.event.TextEvent)
         return false;

      Point loc = gd.getLocation();
      Dimension ws = gd.getSize();
      ImageWindow.setNextLocation(loc.x+ws.width, loc.y+0*ws.height);

      ImagePlus bgimp = null;
      Roi       bgroi = null;
      int bgin = gd.getNextChoiceIndex();
      if (bgin > 0) {
         bgimp = WindowManager.getImage(ids[bgin-1]);
         bgroi = bgimp.getRoi();
         bgimp = new ImagePlus(null,bgimp.getProcessor());
         }

      ImagePlus fgimp = null;
      Roi       fgroi = null;
      int fgin = gd.getNextChoiceIndex();
      if (fgin > 0) {
         fgimp = WindowManager.getImage(ids[fgin-1]);
         fgroi = fgimp.getRoi();
         fgimp = new ImagePlus(null,fgimp.getProcessor());
         }

      if ((bgimp == null) && (fgimp == null))
         return false;

      String rf = gd.getNextString();
      boolean rfq = gd.getNextBoolean();
      if ((rf.equals("") || rfq)  && (fgimp != null)) {
         rf = ""+fgimp.getDisplayRangeMin()+" "+fgimp.getDisplayRangeMax();
         ((TextField)gd.getStringFields().get(0)).setText(rf);
         }
      String[] rfs = rf.split("\\s");
      float fgll=0, fgul=0;
      if (fgimp != null) {
         try { fgll = Float.parseFloat(rfs[0]); } catch(Throwable t) {fgll = (float)fgimp.getDisplayRangeMin(); }
         try { fgul = Float.parseFloat(rfs[1]); } catch(Throwable t) {fgul = (float)fgimp.getDisplayRangeMax(); }
         }

      String lut = gd.getNextChoice();

      newimp = doOverlay(bgimp, bgroi, fgimp, fgroi, fgll, fgul, lut);
      newimp.changes = false;

      String scale         = gd.getNextString();
      String interpolation = gd.getNextChoice();
      String boarder       = gd.getNextChoice();

      try { newimp = doScale(newimp, scale, interpolation, boarder); } catch(Throwable e) {IJ.log("Bad Scale");}
      newimp.changes = false;

      String cbwhere = gd.getNextChoice();
      String cbsize  = gd.getNextString();

      String label  = gd.getNextString();
      double multiplier = 1;
      try { multiplier = Double.parseDouble(gd.getNextString()); } catch(Throwable e) {IJ.log("Bad Multiplier");}
      String format = gd.getNextString();

      String font  = gd.getNextChoice();
      int    style = new int[]{Font.PLAIN,Font.BOLD,Font.ITALIC,Font.BOLD+Font.ITALIC}[gd.getNextChoiceIndex()];
//      int    just  = new int[]{TextRoi.LEFT,TextRoi.CENTER,TextRoi.RIGHT}[gd.getNextChoiceIndex()];
      String fgc   = gd.getNextChoice();
      String bgc   = gd.getNextChoice();
      int    fontSize = (int)gd.getNextNumber();
//      int    textAngle = (int)gd.getNextNumber();
//      boolean aa = gd.getNextBoolean();

      if ((label != null) && !label.equals("")) {
         newimp = addLabel(newimp, 0, 0, label, new Font(font, style, fontSize), Colors.getColor(fgc,null), Colors.getColor(bgc,null));
         newimp.changes = false;
         }

      if (fgimp == null) lut = null;
      newimp = addColorBar(newimp, lut, cbwhere, cbsize, fgll*multiplier, fgul*multiplier, format, new Font(font, style, fontSize), Colors.getColor(fgc,null), Colors.getColor(bgc,null));

      if (newimp.isVisible()) newimp.updateAndRepaintWindow();
                         else newimp.show();
      newimp.changes = false;
      
      if (fgimp != null) fgimp.setRoi(fgroi);
      return true;
      }

   /** scale the image appropriately.
    */
   public static ImagePlus doScale(ImagePlus imp, String newSize, String interpolationType, String boarderType) {
      ImagePlus newimp = imp.duplicate();
      int[] ns = parseSize(newSize, imp.getWidth(), imp.getHeight(), "100%", "100%");
      IJ.run(imp, "Size...", "width="+ns[0]+" weight="+ns[1]+" constrain average interpolation="+interpolationType);
      IJ.run(imp, "Canvas Size...", "width="+(ns[0]+ns[2])+" height="+(ns[1]+ns[3])+" position="+boarderType);
      return imp;
      }

   /** Overlay colorbar and label.
    */
   public static ImagePlus doOverlayLabelColorBar(ImagePlus bgimp, Roi bgroi, ImagePlus fgimp, Roi fgroi, 
                                                  String label, double fgll, double fgul, double multiplier,
                                                  String lut, String cbwhere, String cbsize, String format, Font font, Color fgc, Color bgc) {
      if (Double.isNaN(fgll)) fgll = fgimp.getDisplayRangeMin();
      if (Double.isNaN(fgul)) fgul = fgimp.getDisplayRangeMax();
      ImagePlus newimp = doOverlay(bgimp, bgroi, fgimp, fgroi, fgll, fgul, lut);
      newimp = addLabel(newimp, 0, 0, label, font, fgc, bgc);
      newimp = addColorBar(newimp, lut, cbwhere, cbsize, fgll*multiplier, fgul*multiplier, format, font, fgc, bgc);
      newimp.changes = false;

      return newimp;
      }

   /** Scale then overlay colorbar and label.
    */
   public static ImagePlus doOverlayScaleLabelColorBar(ImagePlus bgimp, Roi bgroi, ImagePlus fgimp, Roi fgroi, 
                                                  String label, double fgll, double fgul, double multiplier, String lut, 
                                                  String newSize, String interpolationType, String boarderType,
                                                  String cbwhere, String cbsize, String format, Font font, Color fgc, Color bgc) {
      if (Double.isNaN(fgll)) fgll = (fgimp!=null?fgimp:bgimp).getDisplayRangeMin();
      if (Double.isNaN(fgul)) fgul = (fgimp!=null?fgimp:bgimp).getDisplayRangeMax();
      ImagePlus newimp = doOverlay(bgimp, bgroi, fgimp, fgroi, fgll, fgul, lut);
      newimp = doScale(newimp, newSize, interpolationType, boarderType);
      newimp = addLabel(newimp, 0, 0, label, font, fgc, bgc);
      newimp = addColorBar(newimp, lut, cbwhere, cbsize, fgll*multiplier, fgul*multiplier, format, font, fgc, bgc);
      newimp.changes = false;

      return newimp;
      }

   /** utility method to get list of Fonts.
    */
   public static String[] getFonts() {
        GraphicsEnvironment ge = GraphicsEnvironment.getLocalGraphicsEnvironment();
        String[] fonts = ge.getAvailableFontFamilyNames();
        ArrayList<String> names = new ArrayList<String>();
        names.add("SansSerif");
        names.add("Serif");
        names.add("Monospaced");
        for (int i=0; i<fonts.length; i++) {
            String f = fonts[i];
            if (f.length()<=20 && !(f.equals("SansSerif")||f.equals("Serif")||f.equals("Monospaced")))
                names.add(f);
        }
        return (String[])names.toArray(new String[names.size()]);
    }

}
